/* OGMRip - A library for DVD ripping and encoding
 * Copyright (C) 2004-2009 Olivier Rolland <billl@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib/gstdio.h>
#include <glib/gi18n-lib.h>

#include "ogmrip.h"
#include "ogmrip-mplayer.h"
#include "ogmjob.h"

#define OGMRIP_TYPE_VP8          (ogmrip_vp8_get_type ())
#define OGMRIP_VP8(obj)          (G_TYPE_CHECK_INSTANCE_CAST ((obj), OGMRIP_TYPE_VP8, OGMRipVp8))
#define OGMRIP_VP8_CLASS(klass)  (G_TYPE_CHECK_CLASS_CAST ((klass), OGMRIP_TYPE_VP8, OGMRipVp8Class))
#define OGMRIP_IS_VP8(obj)       (G_TYPE_CHECK_INSTANCE_TYPE ((obj), OGMRIP_TYPE_VP8))
#define OGMRIP_IS_VP8_CLASS(obj) (G_TYPE_CHECK_CLASS_TYPE ((klass), OGMRIP_TYPE_VP8))

typedef struct _OGMRipVp8      OGMRipVp8;
typedef struct _OGMRipVp8Class OGMRipVp8Class;

struct _OGMRipVp8
{
  OGMRipVideoCodec parent_instance;

  gboolean best;
  gint minsection_pct;
  gint maxsection_pct;
  gint token_parts;
  gint drop_frame;
  gint min_q;
  gint max_q;
  gint profile;
  gint cpu_used;
};

struct _OGMRipVp8Class
{
  OGMRipVideoCodecClass parent_class;
};

static GType ogmrip_vp8_get_type    (void);
static gint  ogmrip_vp8_run         (OGMJobSpawn       *spawn);
static void  ogmrip_vp8_set_quality (OGMRipVideoCodec  *video,
                                     OGMRipQualityType quality);

static gchar **
ogmrip_yuv4mpeg_command (OGMRipVideoCodec *video, const gchar *fifo)
{
  OGMDvdTitle *title;
  GPtrArray *argv;
  gint vid;

  g_return_val_if_fail (OGMRIP_IS_VIDEO_CODEC (video), NULL);
  g_return_val_if_fail (fifo != NULL, NULL);

  title = ogmrip_codec_get_input (OGMRIP_CODEC (video));
  g_return_val_if_fail (title != NULL, NULL);

  argv = ogmrip_mplayer_video_command (video, fifo);

  g_ptr_array_add (argv, g_strdup ("-vo"));
  if (MPLAYER_CHECK_VERSION (1,0,0,6))
    g_ptr_array_add (argv, g_strdup_printf ("yuv4mpeg:file=%s", fifo));
  else
    g_ptr_array_add (argv, g_strdup ("yuv4mpeg"));

  vid = ogmdvd_title_get_nr (title);

  if (MPLAYER_CHECK_VERSION (1,0,0,1))
    g_ptr_array_add (argv, g_strdup_printf ("dvd://%d", vid + 1));
  else
  {
    g_ptr_array_add (argv, g_strdup ("-dvd"));
    g_ptr_array_add (argv, g_strdup_printf ("%d", vid + 1));
  }

  g_ptr_array_add (argv, NULL);

  return (gchar **) g_ptr_array_free (argv, FALSE);
}

static gchar **
ogmrip_vp8_command (OGMRipVideoCodec *video, const gchar *fifo, guint pass, guint passes, const gchar *log_file)
{
  GPtrArray *argv;
  const gchar *output;
  guint width, height, num, denom;
  gint bitrate, threads;

  g_return_val_if_fail (OGMRIP_IS_VIDEO_CODEC (video), NULL);
  g_return_val_if_fail (fifo != NULL, NULL);

  output = ogmrip_codec_get_output (OGMRIP_CODEC (video));
  g_return_val_if_fail (output != NULL, NULL);

  argv = g_ptr_array_new ();

  g_ptr_array_add (argv, g_strdup ("vpxenc"));

  ogmrip_video_codec_get_scale_size (video, &width, &height);
  g_ptr_array_add (argv, g_strdup_printf ("--width=%u", width));
  g_ptr_array_add (argv, g_strdup_printf ("--height=%u", height));

  ogmrip_codec_get_framerate (OGMRIP_CODEC (video), &num, &denom);
  g_ptr_array_add (argv, g_strdup_printf ("--timebase=%u/%u", denom, num));

  bitrate = ogmrip_video_codec_get_bitrate (video);
  if (bitrate > 0)
  {
    g_ptr_array_add (argv, g_strdup_printf ("--target-bitrate=%d", bitrate / 1000));

    if (OGMRIP_VP8 (video)->min_q >= 0)
      g_ptr_array_add (argv, g_strdup_printf ("--min-q=%u", OGMRIP_VP8 (video)->min_q));

    if (OGMRIP_VP8 (video)->max_q >= 0)
      g_ptr_array_add (argv, g_strdup_printf ("--max-q=%u", OGMRIP_VP8 (video)->max_q));
  }
  else
  {
    gdouble quantizer;

    quantizer = 48. * ogmrip_video_codec_get_quantizer (video) / 31.;
    g_ptr_array_add (argv, g_strdup_printf ("--min-q=%.0lf", quantizer));
    g_ptr_array_add (argv, g_strdup_printf ("--max-q=%.0lf", quantizer + 15.));
  }

  g_ptr_array_add (argv, g_strdup_printf ("--passes=%u", passes));

  if (passes > 1 && log_file)
  {
    g_ptr_array_add (argv, g_strdup_printf ("--pass=%u", pass));
    g_ptr_array_add (argv, g_strdup_printf ("--fpf=%s", log_file));

    if (OGMRIP_VP8 (video)->minsection_pct >= 0)
      g_ptr_array_add (argv, g_strdup_printf ("--minsection-pct=%u", OGMRIP_VP8 (video)->minsection_pct));

    if (OGMRIP_VP8 (video)->maxsection_pct >= 0)
      g_ptr_array_add (argv, g_strdup_printf ("--maxsection-pct=%u", OGMRIP_VP8 (video)->maxsection_pct));
  }

  g_ptr_array_add (argv, g_strdup ("--end-usage=0"));
  g_ptr_array_add (argv, g_strdup ("--auto-alt-ref=1"));
  g_ptr_array_add (argv, g_strdup ("--lag-in-frames=16"));
  g_ptr_array_add (argv, g_strdup ("--kf-min-dist=0"));
  g_ptr_array_add (argv, g_strdup ("--kf-max-dist=360"));
  g_ptr_array_add (argv, g_strdup ("--static-thresh=0"));

  g_ptr_array_add (argv, g_strdup (OGMRIP_VP8 (video)->best ? "--best" : "--good"));

  threads = ogmrip_video_codec_get_threads (video);
  if (threads > 0)
    g_ptr_array_add (argv, g_strdup_printf ("--threads=%u", threads));

  if (OGMRIP_VP8 (video)->token_parts >= 0)
    g_ptr_array_add (argv, g_strdup_printf ("--token-parts=%u", OGMRIP_VP8 (video)->token_parts));

  if (OGMRIP_VP8 (video)->drop_frame >= 0)
    g_ptr_array_add (argv, g_strdup_printf ("--drop-frame=%u", OGMRIP_VP8 (video)->drop_frame));

  if (OGMRIP_VP8 (video)->profile >= 0)
    g_ptr_array_add (argv, g_strdup_printf ("--profile=%u", OGMRIP_VP8 (video)->profile));

  if (OGMRIP_VP8 (video)->cpu_used >= 0)
    g_ptr_array_add (argv, g_strdup_printf ("--cpu-used=%u", OGMRIP_VP8 (video)->cpu_used));

  g_ptr_array_add (argv, g_strdup ("-o"));
  g_ptr_array_add (argv, g_strdup (output));

  g_ptr_array_add (argv, g_strdup (fifo));

  g_ptr_array_add (argv, NULL);

  return (gchar **) g_ptr_array_free (argv, FALSE);
}

G_DEFINE_TYPE (OGMRipVp8, ogmrip_vp8, OGMRIP_TYPE_VIDEO_CODEC)

static void
ogmrip_vp8_class_init (OGMRipVp8Class *klass)
{
  OGMJobSpawnClass *spawn_class;
  OGMRipVideoCodecClass *video_class;

  spawn_class = OGMJOB_SPAWN_CLASS (klass);
  spawn_class->run = ogmrip_vp8_run;

  video_class = OGMRIP_VIDEO_CODEC_CLASS (klass);
  video_class->set_quality = ogmrip_vp8_set_quality;
}

static void
ogmrip_vp8_init (OGMRipVp8 *vp8)
{
  vp8->best = FALSE;
  vp8->minsection_pct = -1;
  vp8->maxsection_pct = -1;
  vp8->token_parts = -1;
  vp8->drop_frame = -1;
  vp8->min_q = -1;
  vp8->max_q = -1;
  vp8->profile = -1;
  vp8->cpu_used = -1;
}

static OGMJobSpawn *
ogmrip_vp8_pipeline (OGMJobSpawn *spawn, const gchar *fifo, guint pass, guint passes, const gchar *log_file)
{
  OGMJobSpawn *pipeline, *child;
  gchar **argv;

  pipeline = ogmjob_pipeline_new ();

  argv = ogmrip_yuv4mpeg_command (OGMRIP_VIDEO_CODEC (spawn), fifo);
  if (argv)
  {
    child = ogmjob_exec_newv (argv);
    ogmjob_exec_add_watch_full (OGMJOB_EXEC (child), (OGMJobWatch) ogmrip_mplayer_video_watch, spawn, TRUE, FALSE, FALSE);
    ogmjob_container_add (OGMJOB_CONTAINER (pipeline), child);
    g_object_unref (child);

    argv = ogmrip_vp8_command (OGMRIP_VIDEO_CODEC (spawn), fifo, pass + 1, passes, log_file);
    if (argv)
    {
      child = ogmjob_exec_newv (argv);
      ogmjob_container_add (OGMJOB_CONTAINER (pipeline), child);
      g_object_unref (child);
    }
  }

  return pipeline;
}

static gint
ogmrip_vp8_run (OGMJobSpawn *spawn)
{
  GError *error = NULL;
  OGMJobSpawn *queue, *pipeline;
  gint pass, passes, result;
  gchar *fifo, *log_file;

  result = OGMJOB_RESULT_ERROR;

  fifo = ogmrip_fs_mkftemp ("fifo.XXXXXX", &error);
  if (!fifo)
  {
    ogmjob_spawn_propagate_error (spawn, error);
    return OGMJOB_RESULT_ERROR;
  }

  queue = ogmjob_queue_new ();
  ogmjob_container_add (OGMJOB_CONTAINER (spawn), queue);
  g_object_unref (queue);

  passes = ogmrip_video_codec_get_passes (OGMRIP_VIDEO_CODEC (spawn));

  log_file = passes > 1 ? ogmrip_fs_mktemp ("log.XXXXXX", NULL) : NULL;

  for (pass = 0; pass < passes; pass ++)
  {
    pipeline = ogmrip_vp8_pipeline (spawn, fifo, pass, passes, log_file);
    ogmjob_container_add (OGMJOB_CONTAINER (queue), pipeline);
    g_object_unref (pipeline);
  }

  result = OGMJOB_SPAWN_CLASS (ogmrip_vp8_parent_class)->run (spawn);

  ogmjob_container_remove (OGMJOB_CONTAINER (spawn), queue);

  if (g_file_test (fifo, G_FILE_TEST_EXISTS))
    g_unlink (fifo);
  g_free (fifo);

  if (g_file_test (log_file, G_FILE_TEST_EXISTS))
    g_unlink (log_file);
  g_free (log_file);

  return result;
}

static void
ogmrip_vp8_set_quality (OGMRipVideoCodec *video, OGMRipQualityType quality)
{
  OGMRipVp8 *vp8;

  vp8 = OGMRIP_VP8 (video);

  switch (quality)
  {
    case OGMRIP_QUALITY_EXTREME:
      vp8->best = FALSE;
      vp8->minsection_pct = 5;
      vp8->maxsection_pct = 800;
      vp8->token_parts = 2;
      vp8->drop_frame = 0;
      vp8->min_q = 0;
      vp8->max_q = 60;
      vp8->profile = -1;
      vp8->cpu_used = 0;
      break;
    case OGMRIP_QUALITY_HIGH:
      vp8->best = FALSE;
      vp8->minsection_pct = 5;
      vp8->maxsection_pct = 800;
      vp8->token_parts = 2;
      vp8->drop_frame = -1;
      vp8->min_q = 0;
      vp8->max_q = 60;
      vp8->profile = -1;
      vp8->cpu_used = 1;
      break;
    default:
      vp8->best = FALSE;
      vp8->minsection_pct = 15;
      vp8->maxsection_pct = 400;
      vp8->token_parts = -1;
      vp8->drop_frame = -1;
      vp8->min_q = 4;
      vp8->max_q = 63;
      vp8->profile = 1;
      vp8->cpu_used = 0;
      break;
  }
}

static OGMRipVideoPlugin vp8_plugin =
{
  NULL,
  G_TYPE_NONE,
  "vp8",
  N_("Vp8"),
  OGMRIP_FORMAT_VP8,
  2,
  G_MAXINT
};

OGMRipVideoPlugin *
ogmrip_init_plugin (GError **error)
{
  gchar *filename;

  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  filename = g_find_program_in_path ("vpxenc");
  if (!filename)
  {
    g_set_error (error, OGMRIP_PLUGIN_ERROR, OGMRIP_PLUGIN_ERROR_REQ, _("vpxenc is missing"));
    return NULL;
  }
  g_free (filename);

  vp8_plugin.type = OGMRIP_TYPE_VP8;

  return &vp8_plugin;
}

